@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2013 Junjiro R. Okajima
+ * Copyright (C) 2005-2014 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -63,8 +63,7 @@ int au_wh_name_alloc(struct qstr *wh, const struct qstr *name)
  * test if the @wh_name exists under @h_parent.
  * @try_sio specifies the necessary of super-io.
  */
-int au_wh_test(struct dentry *h_parent, struct qstr *wh_name,
-	       struct au_branch *br, int try_sio)
+int au_wh_test(struct dentry *h_parent, struct qstr *wh_name, int try_sio)
 {
 	int err;
 	struct dentry *wh_dentry;
@@ -72,7 +71,7 @@ int au_wh_test(struct dentry *h_parent, struct qstr *wh_name,
 	if (!try_sio)
 		wh_dentry = vfsub_lkup_one(wh_name, h_parent);
 	else
-		wh_dentry = au_sio_lkup_one(wh_name, h_parent, br);
+		wh_dentry = au_sio_lkup_one(wh_name, h_parent);
 	err = PTR_ERR(wh_dentry);
 	if (IS_ERR(wh_dentry))
 		goto out;
@@ -86,8 +85,8 @@ int au_wh_test(struct dentry *h_parent, struct qstr *wh_name,
 		goto out_wh; /* success */
 
 	err = -EIO;
-	AuIOErr("%.*s Invalid whiteout entry type 0%o.\n",
-		AuDLNPair(wh_dentry), wh_dentry->d_inode->i_mode);
+	AuIOErr("%pd Invalid whiteout entry type 0%o.\n",
+		wh_dentry, wh_dentry->d_inode->i_mode);
 
 out_wh:
 	dput(wh_dentry);
@@ -98,13 +97,13 @@ out:
 /*
  * test if the @h_dentry sets opaque or not.
  */
-int au_diropq_test(struct dentry *h_dentry, struct au_branch *br)
+int au_diropq_test(struct dentry *h_dentry)
 {
 	int err;
 	struct inode *h_dir;
 
 	h_dir = h_dentry->d_inode;
-	err = au_wh_test(h_dentry, &diropq_name, br,
+	err = au_wh_test(h_dentry, &diropq_name,
 			 au_test_h_perm_sio(h_dir, MAY_EXEC));
 	return err;
 }
@@ -148,7 +147,7 @@ struct dentry *au_whtmp_lkup(struct dentry *h_parent, struct au_branch *br,
 	qs.name = name;
 	for (i = 0; i < 3; i++) {
 		sprintf(p, "%.*x", AUFS_WH_TMP_LEN, cnt++);
-		dentry = au_sio_lkup_one(&qs, h_parent, br);
+		dentry = au_sio_lkup_one(&qs, h_parent);
 		if (IS_ERR(dentry) || !dentry->d_inode)
 			goto out_name;
 		dput(dentry);
@@ -175,7 +174,7 @@ int au_whtmp_ren(struct dentry *h_dentry, struct au_branch *br)
 	struct path h_path = {
 		.mnt = au_br_mnt(br)
 	};
-	struct inode *h_dir;
+	struct inode *h_dir, *delegated;
 	struct dentry *h_parent;
 
 	h_parent = h_dentry->d_parent; /* dir inode is locked */
@@ -188,8 +187,14 @@ int au_whtmp_ren(struct dentry *h_dentry, struct au_branch *br)
 		goto out;
 
 	/* under the same dir, no need to lock_rename() */
-	err = vfsub_rename(h_dir, h_dentry, h_dir, &h_path);
+	delegated = NULL;
+	err = vfsub_rename(h_dir, h_dentry, h_dir, &h_path, &delegated);
 	AuTraceErr(err);
+	if (unlikely(err == -EWOULDBLOCK)) {
+		pr_warn("cannot retry for NFSv4 delegation"
+			" for an internal rename\n");
+		iput(delegated);
+	}
 	dput(h_path.dentry);
 
 out:
@@ -204,7 +209,8 @@ out:
 
 static int do_unlink_wh(struct inode *h_dir, struct path *h_path)
 {
-	int force;
+	int err, force;
+	struct inode *delegated;
 
 	/*
 	 * forces superio when the dir has a sticky bit.
@@ -212,7 +218,14 @@ static int do_unlink_wh(struct inode *h_dir, struct path *h_path)
 	 */
 	force = (h_dir->i_mode & S_ISVTX)
 		&& !uid_eq(current_fsuid(), h_path->dentry->d_inode->i_uid);
-	return vfsub_unlink(h_dir, h_path, force);
+	delegated = NULL;
+	err = vfsub_unlink(h_dir, h_path, &delegated, force);
+	if (unlikely(err == -EWOULDBLOCK)) {
+		pr_warn("cannot retry for NFSv4 delegation"
+			" for an internal unlink\n");
+		iput(delegated);
+	}
+	return err;
 }
 
 int au_wh_unlink_dentry(struct inode *h_dir, struct path *h_path,
@@ -258,17 +271,25 @@ static void au_wh_clean(struct inode *h_dir, struct path *whpath,
 			const int isdir)
 {
 	int err;
+	struct inode *delegated;
 
 	if (!whpath->dentry->d_inode)
 		return;
 
 	if (isdir)
 		err = vfsub_rmdir(h_dir, whpath);
-	else
-		err = vfsub_unlink(h_dir, whpath, /*force*/0);
+	else {
+		delegated = NULL;
+		err = vfsub_unlink(h_dir, whpath, &delegated, /*force*/0);
+		if (unlikely(err == -EWOULDBLOCK)) {
+			pr_warn("cannot retry for NFSv4 delegation"
+				" for an internal unlink\n");
+			iput(delegated);
+		}
+	}
 	if (unlikely(err))
-		pr_warn("failed removing %.*s (%d), ignored.\n",
-			AuDLNPair(whpath->dentry), err);
+		pr_warn("failed removing %pd (%d), ignored.\n",
+			whpath->dentry, err);
 }
 
 static int test_linkable(struct dentry *h_root)
@@ -278,8 +299,8 @@ static int test_linkable(struct dentry *h_root)
 	if (h_dir->i_op->link)
 		return 0;
 
-	pr_err("%.*s (%s) doesn't support link(2), use noplink and rw+nolwh\n",
-	       AuDLNPair(h_root), au_sbtype(h_root->d_sb));
+	pr_err("%pd (%s) doesn't support link(2), use noplink and rw+nolwh\n",
+	       h_root, au_sbtype(h_root->d_sb));
 	return -ENOSYS;
 }
 
@@ -298,7 +319,7 @@ static int au_whdir(struct inode *h_dir, struct path *path)
 	} else if (S_ISDIR(path->dentry->d_inode->i_mode))
 		err = 0;
 	else
-		pr_err("unknown %.*s exists\n", AuDLNPair(path->dentry));
+		pr_err("unknown %pd exists\n", path->dentry);
 
 	return err;
 }
@@ -321,9 +342,9 @@ static void au_wh_init_ro(struct inode *h_dir, struct au_wh_base base[],
 
 /*
  * returns tri-state,
- * minus: error, caller should print the mesage
+ * minus: error, caller should print the message
  * zero: succuess
- * plus: error, caller should NOT print the mesage
+ * plus: error, caller should NOT print the message
  */
 static int au_wh_init_rw_nolink(struct dentry *h_root, struct au_wbr *wbr,
 				int do_plink, struct au_wh_base base[],
@@ -393,8 +414,7 @@ static int au_wh_init_rw(struct dentry *h_root, struct au_wbr *wbr,
 	} else if (S_ISREG(base[AuBrWh_BASE].dentry->d_inode->i_mode))
 		err = 0;
 	else
-		pr_err("unknown %.*s/%.*s exists\n",
-		       AuDLNPair(h_root), AuDLNPair(base[AuBrWh_BASE].dentry));
+		pr_err("unknown %pd2 exists\n", base[AuBrWh_BASE].dentry);
 	if (unlikely(err))
 		goto out;
 
@@ -497,8 +517,8 @@ int au_wh_init(struct au_branch *br, struct super_block *sb)
 	goto out; /* success */
 
 out_err:
-	pr_err("an error(%d) on the writable branch %.*s(%s)\n",
-	       err, AuDLNPair(h_root), au_sbtype(h_root->d_sb));
+	pr_err("an error(%d) on the writable branch %pd(%s)\n",
+	       err, h_root, au_sbtype(h_root->d_sb));
 out:
 	for (i = 0; i < AuBrWh_Last; i++)
 		dput(base[i].dentry);
@@ -524,7 +544,7 @@ static void reinit_br_wh(void *arg)
 	struct path h_path;
 	struct reinit_br_wh *a = arg;
 	struct au_wbr *wbr;
-	struct inode *dir;
+	struct inode *dir, *delegated;
 	struct dentry *h_root;
 	struct au_hinode *hdir;
 
@@ -551,10 +571,16 @@ static void reinit_br_wh(void *arg)
 	if (!err) {
 		h_path.dentry = wbr->wbr_whbase;
 		h_path.mnt = au_br_mnt(a->br);
-		err = vfsub_unlink(hdir->hi_inode, &h_path, /*force*/0);
+		delegated = NULL;
+		err = vfsub_unlink(hdir->hi_inode, &h_path, &delegated,
+				   /*force*/0);
+		if (unlikely(err == -EWOULDBLOCK)) {
+			pr_warn("cannot retry for NFSv4 delegation"
+				" for an internal unlink\n");
+			iput(delegated);
+		}
 	} else {
-		pr_warn("%.*s is moved, ignored\n",
-			AuDLNPair(wbr->wbr_whbase));
+		pr_warn("%pd is moved, ignored\n", wbr->wbr_whbase);
 		err = 0;
 	}
 	dput(wbr->wbr_whbase);
@@ -564,6 +590,8 @@ static void reinit_br_wh(void *arg)
 	wbr_wh_write_unlock(wbr);
 	au_hn_imtx_unlock(hdir);
 	di_read_unlock(a->sb->s_root, AuLock_IR);
+	if (!err)
+		au_fhsm_wrote(a->sb, bindex, /*force*/0);
 
 out:
 	if (wbr)
@@ -624,7 +652,7 @@ static int link_or_create_wh(struct super_block *sb, aufs_bindex_t bindex,
 	struct au_branch *br;
 	struct au_wbr *wbr;
 	struct dentry *h_parent;
-	struct inode *h_dir;
+	struct inode *h_dir, *delegated;
 
 	h_parent = wh->d_parent; /* dir inode is locked */
 	h_dir = h_parent->d_inode;
@@ -635,7 +663,13 @@ static int link_or_create_wh(struct super_block *sb, aufs_bindex_t bindex,
 	wbr = br->br_wbr;
 	wbr_wh_read_lock(wbr);
 	if (wbr->wbr_whbase) {
-		err = vfsub_link(wbr->wbr_whbase, h_dir, &h_path);
+		delegated = NULL;
+		err = vfsub_link(wbr->wbr_whbase, h_dir, &h_path, &delegated);
+		if (unlikely(err == -EWOULDBLOCK)) {
+			pr_warn("cannot retry for NFSv4 delegation"
+				" for an internal link\n");
+			iput(delegated);
+		}
 		if (!err || err != -EMLINK)
 			goto out;
 
@@ -645,6 +679,8 @@ static int link_or_create_wh(struct super_block *sb, aufs_bindex_t bindex,
 
 	/* return this error in this context */
 	err = vfsub_create(h_dir, &h_path, WH_MASK, /*want_excl*/true);
+	if (!err)
+		au_fhsm_wrote(sb, bindex, /*force*/0);
 
 out:
 	wbr_wh_read_unlock(wbr);
@@ -769,9 +805,10 @@ struct dentry *au_wh_create(struct dentry *dentry, aufs_bindex_t bindex,
 	wh_dentry = au_wh_lkup(h_parent, &dentry->d_name, au_sbr(sb, bindex));
 	if (!IS_ERR(wh_dentry) && !wh_dentry->d_inode) {
 		err = link_or_create_wh(sb, bindex, wh_dentry);
-		if (!err)
+		if (!err) {
 			au_set_dbwh(dentry, bindex);
-		else {
+			au_fhsm_wrote(sb, bindex, /*force*/0);
+		} else {
 			dput(wh_dentry);
 			wh_dentry = ERR_PTR(err);
 		}
@@ -944,8 +981,7 @@ int au_whtmp_rmdir(struct inode *dir, aufs_bindex_t bindex,
 		return 0; /* success */
 	}
 
-	pr_warn("failed removing %.*s(%d), ignored\n",
-		AuDLNPair(wh_dentry), err);
+	pr_warn("failed removing %pd(%d), ignored\n", wh_dentry, err);
 	return err;
 }
 
@@ -1014,8 +1050,7 @@ void au_whtmp_kick_rmdir(struct inode *dir, aufs_bindex_t bindex,
 	args->wh_dentry = dget(wh_dentry);
 	wkq_err = au_wkq_nowait(call_rmdir_whtmp, args, sb, /*flags*/0);
 	if (unlikely(wkq_err)) {
-		pr_warn("rmdir error %.*s (%d), ignored\n",
-			AuDLNPair(wh_dentry), wkq_err);
+		pr_warn("rmdir error %pd (%d), ignored\n", wh_dentry, wkq_err);
 		au_whtmp_rmdir_free(args);
 	}
 }
